/*                                      (   Q  )
                                           ( L  )
                                             ( T)
                                               |Q|
  \----------------------------------------------------/
  |   Genome scanning using QTLs                       |
  |   Jing Hua Zhao                                    |
  |                                                    |
  |   26-8-1999 drafting                               |
  |   31-8-1999 multistage genome scans                .
  |    4-9-1999 add improved smoothing                 .
  |             see QTLSIM for detailed comments       .
  /----------------------------------------------------\
   O O O O                 O O O O              O O O O
 */
%macro multiHE(H0=0,               /*hypothesis indicator*/
               mapscale=0.1,       /*magnifying factor*/
               seed=123,           /*sib correlation*/
               rs=0.25,            /*random number seed*/
               ssize=3000,         /*sample size*/
);

proc iml;
/*
 * retrieve parameters of simulation
 */
h0=&h0;
mapscale=&mapscale;
rs=&rs;
seed=&seed+&i;
ssize=&ssize;

/*
 * define the genome
 */
morton=
{
 294,266,213,238,211,210,178,172,147,174,150,
 179,130,118,151,151,202,143,147,151,92,81,220
};

chrlen=j(1,23,0);
clen=j(1,23,0);
c=0;
do i=1 to 23;
   chrlen[i]=ceil(morton[i]/mapscale);
   c=c+chrlen[i];
   clen[i]=c;
end;
longest=chrlen[1];
lgenome=chrlen[+];

START seth2;
/*
 * set QTL heritabilities for seven scenarios
 * need h0 in [1,7]
 */
qtl={4,8,16,32,9,10,13};
scene1=j(1,qtl[1],0.1);
scene2=j(1,qtl[2],0.05);
scene3=j(1,qtl[3],0.025);
scene4=j(1,qtl[4],0.0125);
scene5=j(1,qtl[5],0);
scene6=j(1,qtl[6],0);
scene7=j(1,qtl[7],0);

p0=0.10;
do i=1 to qtl[5];
   scene5[i]=p0;
   p0=p0*0.75;
end;
p0=0.075;
do i=1 to qtl[6];
   scene6[i]=p0;
   p0=p0*0.8125;
end;
p0=0.05;
do i=1 to qtl[7];
   scene7[i]=p0;
   p0=p0*0.875;
end;

s1=scene1[+];
s2=scene2[+];
s3=scene3[+];
s4=scene4[+];
s5=scene5[+];
s6=scene6[+];
s7=scene7[+];

* print scene1 scene2 scene3 scene4 scene5 scene6 scene7;
* print s1 s2 s3 s4 s5 s6 s7;

tqtl=qtl[h0];

     if h0=1 then h2qtl=scene1;
else if h0=2 then h2qtl=scene2;
else if h0=3 then h2qtl=scene3;
else if h0=4 then h2qtl=scene4;
else if h0=5 then h2qtl=scene5;
else if h0=6 then h2qtl=scene6;
else if h0=7 then h2qtl=scene7;
else do;
     tqtl=qtl[1];
     h2qtl=scene1;
end;

FINISH seth2;

run seth2;
/*
 * position the QTLs
 */
chrpos=j(tqtl,23,0);
pos=j(2,tqtl,0);
pctqtl=j(1,tqtl,0);
nqtl=j(1,23,0);
VA=j(tqtl,23,0);
pchr=clen/lgenome;
do i=1 to tqtl;
   j=1;
   x=ranuni(seed);
   do k=1 to 23;
      if x>=pchr[k] then j=j+1;
   end;
   nqtl[j]=nqtl[j]+1;
   if j=1 then c=1+(chrlen[j]-1)*x;
   else c=1+(lgenome-1)*x-clen[j-1];
   c=ceil(c);
   chrpos[nqtl[j],j]=c;
   pos[1,i]=j;
   pos[2,i]=c;
   VA[nqtl[j],j]=h2qtl[i];
end;

START hinv(x);
      return (0.5*(1-exp(-2*abs(x))));
FINISH hinv;

/*
 * prepare weights
 */
workwi=j(longest,1,0);
do i=1 to longest;
   d=i/100*mapscale;
   theta=hinv(d);
   workwi[i]=(1-2*theta)**2;
end;

START onescan;
/*
 * critical procedure
 */
beta=j(longest,23,0);
lod=beta;
sys=beta;
ran=beta;
VR=beta;

c0=8*(1+rs**2)/ssize;
do i=1 to 23;
   points=chrlen[i];
   do k=1 to nqtl[i];
      do j=1 to points;
         d=abs(j-chrpos[k,i]);
         if d=0 then c=1;
         else c=workwi[d];
         if h0=0 then sys[j,i]=0;
         else sys[j,i]=sys[j,i]+c*VA[k,i];
      end;
   end;
   do j=1 to points;
      VR[j,i]=c0-sys[j,i]**2/ssize;
   end;
   d=1/100*mapscale;
   theta=hinv(d);
   cc=(1-2*theta)**2;
   rani=sqrt(VR[1,i])*rannor(seed);
   beta[1,i]=sys[1,i]+rani;
   do j=2 to points;
      vi=VR[j,i];
      vi1=VR[j-1,i];
      c2=cc*sqrt(vi/vi1);
      d2=vi-cc*sqrt(vi*vi1);
      rani=sqrt(c2)*rani+sqrt(d2)*rannor(seed);
      ran[j,i]=rani;
      beta[j,i]=sys[j,i]+rani;
   end;
   do j=1 to points;
      if beta[j,i]>0 then lod[j,i]=beta[j,i]**2/2/log(10)/c0;
   end;
end;
free vr;

FINISH onescan;

/*
 * 1 - generate stages (ssize) = 1(3,000), 2(2,000), 3(2,000)
 */
ssize=3000;
run onescan;
sys1=sys;
ran1=ran;

ssize=2000;
run onescan;
ran2=ran;

ssize=2000;
run onescan;
ran3=ran;
/*
 * 2 - obtain the stage 1 top 30% (well, 100%)
 */
/*
 * rank betas over all points
 */
allpoint=j(lgenome,1,0);
k=1;
do i=1 to 23;
   do j=1 to chrlen[i];
      allpoint[k]=beta[j,i];
      k=k+1;
   end;
end;
pctall=j(longest,23,0);
k=1;
ranks=rank(allpoint);
do i=1 to 23;
   do j=1 to chrlen[i];
      pctall[j,i]=ranks[k]/lgenome*100;
      k=k+1;
   end;
end;
free allpoint ranks;

/*
 * 3 - combine stages 1(30%) and 2
 *     smooth over betas of stages 1 and 2
 */
beta=j(longest,23,0);
do i=1 to 23;
   points=chrlen[i];
   do j=1 to points;
      beta[j,i]=(ran1[i,i]*3+ran2[j,i]*2)/5;
      if pctall[j,i]>70 then do;
         beta[j,i]=beta[j,i]+sys1[j,i];
      end;
   end;
end;

/*
 * rank betas over 30% points
 */
l30=ceil(lgenome*0.3);
allpoint=j(l30,1,0);
k=1;
do i=1 to 23;
   do j=1 to chrlen[i];
      if pctall[j,i]>70 then do;
         allpoint[k]=beta[j,i];
         k=k+1;
      end;
   end;
end;
pct10=j(longest,23,0);
ranks=rank(allpoint);
k=1;
do i=1 to 23;
   do j=1 to chrlen[i];
      if pctall[j,i]>70 then do;
         pct10[j,i]=ranks[k]/l30*100;
         k=k+1;
      end;
   end;
end;
free allpoint ranks;

/*
 * 4 - calculate betas from stages 10%(1, 2) and 3
 *     smooth over 10% betas
 */
beta=j(longest,23,1);
do i=1 to 23;
   k1=1;
   k2=1;
   points=chrlen[i];
   do j=1 to points;
      beta[j,i]=(3*ran1[j,i]+2*ran2[j,i]+2*ran3[j,i])/7;
      if pct10[j,i]>90 then do;
         beta[j,i]=beta[j,i]+sys1[j,i];
      end;
   end;
end;
free chrlen10;

/*
 * rank betas over 10% points, re-creates pctall
 */
l10=ceil(lgenome*0.1);
allpoint=j(l10,1,0);
k=1;
do i=1 to 23;
   do j=1 to chrlen[i];
      if pct10[j,i]>90 then do;
         allpoint[k]=beta[j,i];
         k=k+1;
      end;
   end;
end;
pctall=j(longest,23,0);
k=1;
ranks=rank(allpoint);
do i=1 to 23;
   do j=1 to chrlen[i];
      if pct10[j,i]>90 then do;
         pctall[j,i]=ranks[k]/l10*100;
         k=k+1;
      end;
   end;
end;
free allpoint ranks;

/*
 * 5 - get %QTL distribution
 */
do i=1 to tqtl;
   k1=pos[2,i];
   k2=pos[1,i];
   x=pctall[k1,k2];
   if x<>0 then pctqtl[i]=100-x;
   else pctqtl[i]=10;
end;

create single from pctqtl;
append from pctqtl;
close single;

quit;

%mend multiHE;

%macro analyse;
data _last_;
  set _last_;
  array qtl _numeric_;
  p1=0;p2=0;p3=0;p4=0;p5=0;
  do over qtl;
     if qtl<=1 then p1=p1+1;
     if qtl<=2 then p2=p2+1;
     if qtl<=3 then p3=p3+1;
     if qtl<=4 then p4=p4+1;
     if qtl<=5 then p5=p5+1;
  end;
run;
proc print;
run;
proc means n mean std;
run;
proc freq;
  table p1-p5;
run;
%mend;

libname x '.';
title;
options nocenter nodate nonumber ps=500 ls=256;
filename logfile '/dev/null';
proc printto log=log;
run;
%macro runall;
%do i=1 %to 1000;
    %multiHE(h0=1,seed=11111);
    proc append base=set1 data=single;
    run;
%end;
%analyse;
data x.set1;
  set set1;
run;
%do i=1 %to 1000;
    %multiHE(h0=2,seed=22222);
    proc append base=set2 data=single;
    run;
%end;
%analyse;
data x.set2;
  set set2;
run;
%do i=1 %to 1000;
    %multiHE(h0=3,seed=33333);
    proc append base=set3 data=single;
    run;
%end;
%analyse;
data x.set3;
  set set3;
run;
%do i=1 %to 1000;
    %multiHE(h0=4,seed=44444);
    proc append base=set4 data=single;
    run;
%end;
%analyse;
data x.set4;
  set set4;
run;
%do i=1 %to 1000;
    %multiHE(h0=5,seed=55555);
    proc append base=set5 data=single;
    run;
%end;
%analyse;
data x.set5;
  set set5;
run;
%do i=1 %to 1000;
    %multiHE(h0=6,seed=66666);
    proc append base=set6 data=single;
    run;
%end;
%analyse;
data x.set6;
  set set6;
run;
%do i=1 %to 1000;
    %multiHE(h0=7,seed=77777);
    proc append base=set7 data=single;
    run;
%end;
%analyse;
data x.set7;
  set set7;
run;
%mend runall;

%runall;
